import functools
import string

import hypothesis.strategies as st
from hypothesis import given

from safe_rcm.product import utils


def shared(*, key):
    def outer(func):
        @functools.wraps(func)
        def inner(*args, **kwargs):
            result = func(*args, **kwargs)

            return st.shared(result, key=key)

        return inner

    return outer


markers = st.characters()
marker = st.shared(markers, key="marker")


def marked_mapping(marker):
    values = st.just(None)

    unmarked_keys = st.text()
    marked_keys = st.builds(lambda k, m: m + k, unmarked_keys, marker)
    keys = st.one_of(unmarked_keys, marked_keys)

    return st.dictionaries(keys, values)


@given(marked_mapping(marker), marker)
def test_split_marked(mapping, marker):
    marked, unmarked = utils.split_marked(mapping, marker=marker)

    assert list(unmarked) == [key for key in mapping if not key.startswith(marker)]


@shared(key="namespaces")
def namespaces():
    values = st.just(None)

    keys = st.text(string.ascii_letters, min_size=1, max_size=4)
    return st.dictionaries(keys, values)


@st.composite
def draw_from(draw, elements):
    elements = draw(elements)

    if not elements:
        return ""

    return draw(st.sampled_from(elements))


def prefixed_names(namespaces):
    def builder(base, prefix):
        return f"{prefix}:{base}" if prefix != "" else base

    bases = st.text(string.ascii_letters, min_size=1)
    all_prefixes = namespaces.map(list)
    prefixes = draw_from(all_prefixes)

    return st.builds(builder, bases, prefixes)


@given(prefixed_names(namespaces()), namespaces())
def test_strip_namespaces(name, namespaces):
    stripped = utils.strip_namespaces(name, namespaces)

    assert ":" not in stripped



import posixpath

import datatree
import numpy as np
import xarray as xr
from tlz.dicttoolz import itemmap, merge_with, valfilter, valmap
from tlz.functoolz import compose_left, curry, flip
from tlz.itertoolz import first

from safe_rcm.product.reader import execute

from .product.dicttoolz import keysplit
from .product.transformers import extract_dataset
from .xml import read_xml


def move_attrs_to_coords(ds, names):
    coords, attrs = keysplit(lambda k: k in names, ds.attrs)

    new = ds.copy()
    new.attrs = attrs

    return new.assign_coords(coords)


def pad_common(dss):
    def compute_padding(item, maximum):
        key, value = item

        return key, (0, maximum[key] - value)

    sizes = [dict(ds.sizes) for ds in dss]
    maximum_sizes = valmap(max, merge_with(list, *sizes))

    pad_widths = [itemmap(flip(compute_padding, maximum_sizes), _) for _ in sizes]

    return [
        ds.pad(padding, mode="constant", constant_values=np.nan)
        for ds, padding in zip(dss, pad_widths)
    ]


def _read_level(mapping):
    return (
        extract_dataset(mapping)
        .pipe(
            lambda ds: ds.swap_dims(
                {first(valfilter(lambda v: v > 1, ds.sizes)): "coefficients"}
            )
        )
        .pipe(lambda ds: ds.reset_coords())
        .pipe(
            move_attrs_to_coords,
            ["sarCalibrationType", "pixelFirstNoiseValue", "stepSize"],
        )
    )


def read_noise_level_file(mapper, path):
    layout = {
        "/referenceNoiseLevel": {
            "path": "/referenceNoiseLevel",
            "f": compose_left(
                curry(map, _read_level),
                curry(map, lambda ds: ds.expand_dims("sarCalibrationType")),
                list,
                curry(xr.combine_by_coords, combine_attrs="drop_conflicts"),
            ),
        },
        "/perBeamReferenceNoiseLevel": {
            "path": "/perBeamReferenceNoiseLevel",
            "f": compose_left(
                curry(map, _read_level),
                curry(map, lambda ds: ds.expand_dims("sarCalibrationType")),
                list,
                pad_common,
                curry(xr.combine_by_coords, combine_attrs="drop_conflicts"),
            ),
        },
        "/azimuthNoiseLevelScaling": {
            "path": "/azimuthNoiseLevelScaling",
            "f": compose_left(
                curry(map, _read_level),
                list,
                pad_common,
                curry(xr.combine_by_coords, combine_attrs="drop_conflicts"),
            ),
        },
    }

    decoded = read_xml(mapper, path)

    converted = valmap(lambda x: execute(**x)(decoded), layout)

    return converted


def read_noise_levels(mapper, root, fnames):
    fnames = fnames.data.tolist()
    paths = [posixpath.join(root, name) for name in fnames]

    poles = [path.removesuffix(".xml").split("_")[1] for path in paths]
    trees = [read_noise_level_file(mapper, path) for path in paths]
    merged = merge_with(list, *trees)
    combined = valmap(
        compose_left(
            curry(xr.concat, dim="pole", combine_attrs="no_conflicts"),
            lambda x: x.assign_coords(pole=poles),
        ),
        merged,
    )

    return datatree.DataTree.from_dict(combined)



from tlz.dicttoolz import get_in
from tlz.itertoolz import first, groupby


def query(path, mapping):
    if path == "/":
        return mapping

    keys = path.lstrip("/").split("/")
    return get_in(keys, mapping, no_default=True)


def itemsplit(predicate, d):
    groups = groupby(predicate, d.items())
    first = dict(groups.get(True, ()))
    second = dict(groups.get(False, ()))

    return first, second


def valsplit(predicate, d):
    wrapper = lambda item: predicate(item[1])
    return itemsplit(wrapper, d)


def keysplit(predicate, d):
    wrapper = lambda item: predicate(item[0])
    return itemsplit(wrapper, d)


def first_values(d):
    return first(d.values())



import numpy as np
from tlz.functoolz import compose, juxt
from tlz.itertoolz import isiterable


def disjunction(*predicates):
    return compose(any, juxt(predicates))


def conjunction(*predicates):
    return compose(all, juxt(predicates))


def is_scalar(x):
    return not isiterable(x) or isinstance(x, (str, bytes))


def is_composite_value(obj):
    if not isinstance(obj, list) or len(obj) not in [1, 2]:
        return False

    if any(not isinstance(el, dict) or list(el) != ["@dataStream", "$"] for el in obj):
        return False

    data_stream_values = [el["@dataStream"].lower() for el in obj]
    return data_stream_values in (["real", "imaginary"], ["magnitude"])


def is_complex(obj):
    return is_composite_value(obj) and len(obj) == 2


def is_magnitude(obj):
    return is_composite_value(obj) and len(obj) == 1


def is_array(obj):
    # definition of a array:
    # - list of scalars
    # - list of 1d lists
    # - complex array:
    #   - complex parts
    #   - list of complex values
    if not isinstance(obj, list):
        return False

    if len(obj) == 0:
        # zero-sized list, not sure what to do here
        return False

    elem = obj[0]
    if is_complex(obj):
        return not is_scalar(elem["$"])
    elif is_scalar(elem):
        return True
    elif isinstance(elem, list):
        if len(elem) == 1 and is_scalar(elem[0]):
            return True
        elif is_complex(elem):
            # array of imaginary values
            return True
        elif all(map(is_scalar, elem)):
            return True

    return False


def is_scalar_variable(obj):
    if not isinstance(obj, dict):
        return False

    if not all(is_scalar(v) for v in obj.values()):
        return False

    return all(k == "$" or k.startswith("@") for k in obj)


is_scalar_valued = disjunction(
    is_scalar, lambda x: is_array(x) and len(x) == 1, is_scalar_variable
)


def is_nested(obj):
    """nested means: list of dict, but all dict values are scalar or 1-valued"""
    if not isinstance(obj, list) or len(obj) == 0:
        return False

    elem = obj[0]
    if not isinstance(elem, dict):
        return False

    if all(map(is_scalar_valued, elem.values())):
        return True

    return False


def is_nested_array(obj):
    return is_nested(obj) and "$" in obj[0]


def is_nested_dataset(obj):
    return is_nested(obj) and "$" not in obj[0]


def is_attr(column):
    """an attribute is a index if it has multiple unique values"""
    return np.unique(column).size == 1



import datatree
import xarray as xr
from tlz.dicttoolz import keyfilter, merge, merge_with, valfilter, valmap
from tlz.functoolz import compose_left, curry, juxt
from tlz.itertoolz import first, second

from ..xml import read_xml
from . import transformers
from .dicttoolz import keysplit, query
from .predicates import disjunction, is_nested_array, is_scalar_valued
from .utils import dictfirst, starcall


@curry
def attach_path(obj, path):
    if not hasattr(obj, "encoding"):
        raise ValueError(
            "cannot attach source path: `obj` does not have a `encoding` attribute."
        )

    new = obj.copy()
    new.encoding["xpath"] = path

    return new


@curry
def execute(mapping, f, path):
    subset = query(path, mapping)

    return compose_left(f, attach_path(path=path))(subset)


def read_product(mapper, product_path):
    decoded = read_xml(mapper, product_path)

    layout = {
        "/": {
            "path": "/",
            "f": curry(transformers.extract_metadata)(collapse=["securityAttributes"]),
        },
        "/sourceAttributes": {
            "path": "/sourceAttributes",
            "f": transformers.extract_metadata,
        },
        "/sourceAttributes/radarParameters": {
            "path": "/sourceAttributes/radarParameters",
            "f": transformers.extract_dataset,
        },
        "/sourceAttributes/radarParameters/prfInformation": {
            "path": "/sourceAttributes/radarParameters/prfInformation",
            "f": transformers.extract_nested_dataset,
        },
        "/sourceAttributes/orbitAndAttitude/orbitInformation": {
            "path": "/sourceAttributes/orbitAndAttitude/orbitInformation",
            "f": compose_left(
                curry(transformers.extract_dataset)(dims="timeStamp"),
                lambda ds: ds.assign_coords(
                    {"timeStamp": ds["timeStamp"].astype("datetime64")}
                ),
            ),
        },
        "/sourceAttributes/orbitAndAttitude/attitudeInformation": {
            "path": "/sourceAttributes/orbitAndAttitude/attitudeInformation",
            "f": compose_left(
                curry(transformers.extract_dataset)(dims="timeStamp"),
                lambda ds: ds.assign_coords(
                    {"timeStamp": ds["timeStamp"].astype("datetime64")}
                ),
            ),
        },
        "/sourceAttributes/rawDataAttributes": {
            "path": "/sourceAttributes/rawDataAttributes",
            "f": compose_left(
                curry(keysplit, lambda k: k != "rawDataAnalysis"),
                juxt(
                    compose_left(first, transformers.extract_dataset),
                    compose_left(
                        second,
                        dictfirst,
                        curry(starcall, curry(merge_with, list)),
                        curry(
                            transformers.extract_dataset,
                            dims={"rawDataHistogram": ["stacked", "histogram"]},
                            default_dims=["stacked"],
                        ),
                        lambda obj: obj.set_index({"stacked": ["pole", "beam"]}),
                        lambda obj: obj.unstack("stacked"),
                    ),
                ),
                curry(xr.merge),
            ),
        },
        "/imageGenerationParameters/generalProcessingInformation": {
            "path": "/imageGenerationParameters/generalProcessingInformation",
            "f": transformers.extract_metadata,
        },
        "/imageGenerationParameters/sarProcessingInformation": {
            "path": "/imageGenerationParameters/sarProcessingInformation",
            "f": compose_left(
                curry(keyfilter, lambda k: k not in {"azimuthWindow", "rangeWindow"}),
                transformers.extract_dataset,
            ),
        },
        "/imageGenerationParameters/chirps": {
            "path": "/imageGenerationParameters/chirp",
            "f": compose_left(
                lambda el: merge_with(list, *el),
                curry(keysplit, lambda k: k != "chirpQuality"),
                juxt(
                    first,
                    compose_left(
                        second,
                        dictfirst,
                        lambda el: merge_with(list, *el),
                    ),
                ),
                lambda x: merge(*x),
                curry(
                    transformers.extract_dataset,
                    dims={
                        "amplitudeCoefficients": ["stacked", "coefficients"],
                        "phaseCoefficients": ["stacked", "coefficients"],
                    },
                    default_dims=["stacked"],
                ),
                lambda obj: obj.set_index({"stacked": ["pole", "pulse"]}),
                lambda obj: obj.drop_duplicates("stacked", keep="last"),
                lambda obj: obj.unstack("stacked"),
            ),
        },
        "/imageGenerationParameters/slantRangeToGroundRange": {
            "path": "/imageGenerationParameters/slantRangeToGroundRange",
            "f": compose_left(
                lambda el: merge_with(list, *el),
                curry(
                    transformers.extract_dataset,
                    dims={
                        "groundToSlantRangeCoefficients": [
                            "zeroDopplerAzimuthTime",
                            "coefficients",
                        ],
                    },
                    default_dims=["zeroDopplerAzimuthTime"],
                ),
            ),
        },
        "/imageReferenceAttributes": {
            "path": "/imageReferenceAttributes",
            "f": compose_left(
                curry(valfilter)(disjunction(is_scalar_valued, is_nested_array)),
                transformers.extract_dataset,
            ),
        },
        "/imageReferenceAttributes/rasterAttributes": {
            "path": "/imageReferenceAttributes/rasterAttributes",
            "f": transformers.extract_dataset,
        },
        "/imageReferenceAttributes/geographicInformation/ellipsoidParameters": {
            "path": "/imageReferenceAttributes/geographicInformation/ellipsoidParameters",
            "f": curry(transformers.extract_dataset)(dims="params"),
        },
        "/imageReferenceAttributes/geographicInformation/geolocationGrid": {
            "path": "/imageReferenceAttributes/geographicInformation/geolocationGrid/imageTiePoint",
            "f": compose_left(
                curry(transformers.extract_nested_datatree)(dims="tie_points"),
                lambda tree: xr.merge([node.ds for node in tree.subtree]),
                lambda ds: ds.set_index(tie_points=["line", "pixel"]),
                lambda ds: ds.unstack("tie_points"),
            ),
        },
        "/imageReferenceAttributes/geographicInformation/rationalFunctions": {
            "path": "/imageReferenceAttributes/geographicInformation/rationalFunctions",
            "f": curry(transformers.extract_dataset)(dims="coefficients"),
        },
        "/sceneAttributes": {
            "path": "/sceneAttributes/imageAttributes",
            "f": compose_left(
                first,  # GRD datasets only have 1
                curry(keyfilter)(lambda x: not x.startswith("@")),
                transformers.extract_dataset,
            ),
        },
        "/grdBurstMap": {
            "path": "/grdBurstMap",
            "f": compose_left(
                curry(
                    map,
                    compose_left(
                        curry(keysplit, lambda k: k != "burstAttributes"),
                        juxt(
                            first,
                            compose_left(
                                second,
                                dictfirst,
                                curry(starcall, curry(merge_with, list)),
                            ),
                        ),
                        curry(starcall, merge),
                        curry(
                            transformers.extract_dataset,
                            dims=["stacked"],
                        ),
                        lambda obj: obj.set_index({"stacked": ["burst", "beam"]}),
                        lambda obj: obj.unstack("stacked"),
                    ),
                ),
                list,
                curry(xr.concat, dim="burst_maps"),
            ),
        },
        "/dopplerCentroid": {
            "path": "/dopplerCentroid",
            "f": compose_left(
                curry(
                    map,
                    compose_left(
                        curry(keysplit, lambda k: k != "dopplerCentroidEstimate"),
                        juxt(
                            first,
                            compose_left(
                                second,
                                dictfirst,
                                curry(starcall, curry(merge_with, list)),
                            ),
                        ),
                        curry(starcall, merge),
                        curry(
                            transformers.extract_dataset,
                            dims={
                                "dopplerCentroidCoefficients": [
                                    "burst",
                                    "coefficients",
                                ],
                            },
                            default_dims=["burst"],
                        ),
                    ),
                ),
                list,
                curry(xr.concat, dim="burst_maps"),
            ),
        },
        "/dopplerRate": {
            "path": "/dopplerRate",
            "f": compose_left(
                curry(
                    map,
                    compose_left(
                        curry(keysplit, lambda k: k != "dopplerRateEstimate"),
                        juxt(
                            first,
                            compose_left(
                                second,
                                dictfirst,
                                curry(starcall, curry(merge_with, list)),
                            ),
                        ),
                        curry(starcall, merge),
                        curry(
                            transformers.extract_dataset,
                            dims={
                                "dopplerRateCoefficients": ["burst", "coefficients"],
                            },
                            default_dims=["burst"],
                        ),
                    ),
                ),
                list,
                curry(xr.concat, dim="burst_maps"),
            ),
        },
    }

    converted = valmap(
        lambda x: execute(**x)(decoded),
        layout,
    )
    return datatree.DataTree.from_dict(converted)



from tlz.functoolz import flip, pipe
from tlz.itertoolz import first, groupby


def split_marked(mapping, marker="@"):
    groups = groupby(lambda item: item[0].startswith(marker), mapping.items())

    attrs = {key.lstrip(marker): value for key, value in groups.get(True, {})}
    data = {key: value for key, value in groups.get(False, {})}

    return attrs, data


def strip_namespaces(name, namespaces):
    """remove the given namespaces from a name

    Parameters
    ----------
    name : str
        The string to trim
    namespaces : sequence of str
        The list of namespaces.

    Returns
    -------
    trimmed : str
        The string without prefix and without leading colon.
    """
    funcs = [flip(str.removeprefix, ns) for ns in namespaces]
    return pipe(name, *funcs).lstrip(":")


def starcall(func, args, **kwargs):
    return func(*args, **kwargs)


def dictfirst(mapping):
    return first(mapping.values())



import datatree
import numpy as np
import xarray as xr
from tlz.dicttoolz import (
    itemfilter,
    itemmap,
    keyfilter,
    keymap,
    merge_with,
    valfilter,
    valmap,
)
from tlz.functoolz import compose_left, curry, flip
from tlz.itertoolz import concat, first, second

from .dicttoolz import first_values, keysplit, valsplit
from .predicates import (
    is_array,
    is_attr,
    is_composite_value,
    is_nested_array,
    is_nested_dataset,
    is_scalar,
)

ignore = ("@xmlns", "@xmlns:xsi", "@xsi:schemaLocation")


def convert_composite(value):
    if not is_composite_value(value):
        raise ValueError(f"not a composite: {value}")

    converted = {part["@dataStream"].lower(): np.array(part["$"]) for part in value}

    if list(converted) == ["magnitude"]:
        return "magnitude", converted["magnitude"]
    else:
        return "complex", converted["real"] + 1j * converted["imaginary"]


def extract_metadata(
    mapping,
    collapse=(),
    ignore=ignore,
):
    without_ignores = keyfilter(lambda k: k not in ignore, mapping)
    # extract the metadata
    metadata_ = itemfilter(
        lambda it: it[0].startswith("@") or is_scalar(it[1]),
        without_ignores,
    )
    metadata = keymap(flip(str.lstrip, "@"), metadata_)

    # collapse the selected items
    to_collapse = keyfilter(lambda x: x in collapse, mapping)
    collapsed = dict(concat(v.items() for v in to_collapse.values()))

    attrs = metadata | collapsed
    return xr.Dataset(attrs=attrs)  # return dataset to avoid bug in datatree


def extract_array(obj, dims):
    if isinstance(dims, str):
        dims = [dims]

    # special case for pulses:
    if "pulses" in dims and len(obj) == 1 and isinstance(obj[0], str):
        obj = obj[0].split()
    elif len(obj) >= 1 and is_composite_value(obj[0]):
        obj = list(map(compose_left(convert_composite, second), obj))
    data = np.array(obj)
    if data.size > 1:
        data = np.squeeze(data)
    return xr.Variable(dims, data)


def extract_composite(obj, dims=()):
    type_, value = convert_composite(obj)

    if is_scalar(value):
        dims = ()
    return xr.Variable(dims, value, {"type": type_})


def extract_variable(obj, dims=()):
    attributes, data = keysplit(lambda k: k.startswith("@"), obj)
    if list(data) != ["$"]:
        raise ValueError("not a variable")

    values = data["$"]
    if is_scalar(values):
        dims = ()

    attrs = keymap(lambda k: k.lstrip("@"), attributes)

    return xr.Variable(dims, values, attrs)


def extract_entry(name, obj, dims=None, default_dims=None):
    if default_dims is None:
        default_dims = [name]

    if isinstance(dims, dict):
        dims = dims.get(name, default_dims)
    elif dims is None:
        dims = default_dims

    if is_array(obj):
        # dimension coordinate
        return extract_array(obj, dims=dims)
    elif is_composite_value(obj):
        return extract_composite(obj, dims=dims)
    elif isinstance(obj, dict):
        return extract_variable(obj, dims=dims)
    elif is_nested_array(obj):
        return extract_nested_array(obj, dims=dims).pipe(rename, name)
    else:
        raise ValueError(f"unknown datastructure:\n{obj}")


def extract_dataset(obj, dims=None, default_dims=None):
    filtered = keyfilter(lambda x: x not in ignore, obj)
    attrs, variables = valsplit(is_scalar, filtered)
    if len(variables) == 1 and is_nested_dataset(first_values(variables)):
        return extract_nested_dataset(first_values(variables), dims=dims).assign_attrs(
            attrs
        )

    variables_ = keymap(lambda k: k.lstrip("@"), variables)

    filtered_variables = valfilter(lambda x: not is_nested_dataset(x), variables_)

    data_vars = itemmap(
        lambda item: (
            item[0],
            extract_entry(*item, dims=dims, default_dims=default_dims),
        ),
        filtered_variables,
    )
    return xr.Dataset(data_vars=data_vars, attrs=attrs)


def extract_nested_variable(obj, dims=None):
    if is_array(obj):
        return xr.Variable(dims, obj)

    columns = merge_with(list, *obj)
    attributes, data = keysplit(lambda k: k.startswith("@"), columns)
    renamed = keymap(lambda k: k.lstrip("@"), attributes)
    attrs = valmap(first, renamed)

    return xr.Variable(dims, data["$"], attrs)


def unstack(obj, dim="stacked"):
    if dim not in obj.dims:
        return obj

    stacked_coords = [name for name, arr in obj.coords.items() if dim in arr.dims]

    return obj.set_index({dim: stacked_coords}).unstack(dim)


def rename(obj, name):
    renamed = obj.rename(name)
    if "$" not in obj.dims:
        return renamed

    if len(obj.dims) != 1:
        raise ValueError(f"unexpected number of dimensions: {list(obj.dims)}")

    return renamed.swap_dims({"$": name})


def to_variable_tuple(name, value, dims):
    if name in dims:
        dims_ = [name]
    else:
        dims_ = dims

    return (dims_, value)


def extract_nested_array(obj, dims=None):
    columns = merge_with(list, *obj)

    attributes, data = keysplit(flip(str.startswith, "@"), columns)
    renamed = keymap(flip(str.lstrip, "@"), attributes)
    preprocessed_attrs = valmap(np.squeeze, renamed)
    attrs_, indexes = valsplit(is_attr, preprocessed_attrs)
    preprocessed_data = valmap(np.squeeze, data)

    originally_stacked = isinstance(dims, (tuple, list)) and "stacked" in dims

    if len(indexes) == 1:
        dims = list(indexes)
    elif len(indexes) >= 2:
        dims = ["stacked"]
    elif dims is None:
        dims = ["$"]

    coords = itemmap(
        lambda it: (it[0], to_variable_tuple(*it, dims=dims)),
        indexes,
    )

    arr = xr.DataArray(
        data=preprocessed_data["$"],
        attrs=valmap(first, attrs_),
        dims=dims,
        coords=coords,
    )
    if originally_stacked:
        return arr
    return arr.pipe(unstack, dim="stacked")


def extract_nested_dataset(obj, dims=None):
    if not isinstance(obj, list):
        raise ValueError(f"unknown type: {type(obj)}")

    columns = merge_with(list, *obj)

    attributes, data = keysplit(flip(str.startswith, "@"), columns)
    renamed = keymap(flip(str.lstrip, "@"), attributes)
    preprocessed = valmap(np.squeeze, renamed)
    attrs_, indexes = valsplit(is_attr, preprocessed)

    attrs = valmap(first, attrs_)

    if dims is None:
        if len(indexes) <= 1:
            dims = list(indexes)
        else:
            dims = ["stacked"]

    data_vars = valmap(curry(extract_nested_variable)(dims=dims), data)
    coords = itemmap(
        lambda it: (it[0], to_variable_tuple(*it, dims=dims)),
        indexes,
    )

    return xr.Dataset(data_vars=data_vars, coords=coords, attrs=attrs).pipe(
        unstack, dim="stacked"
    )


def extract_nested_datatree(obj, dims=None):
    if not isinstance(obj, list):
        raise ValueError(f"unknown type: {type(obj)}")

    datasets = merge_with(list, *obj)
    tree = valmap(curry(extract_nested_dataset)(dims=dims), datasets)

    return datatree.DataTree.from_dict(tree)



import io
import posixpath
import re
from collections import deque

import xmlschema
from lxml import etree
from tlz.dicttoolz import keymap

include_re = re.compile(r'\s*<xsd:include schemaLocation="(?P<location>[^"/]+)"\s?/>')


def remove_includes(text):
    return io.StringIO(include_re.sub("", text))


def extract_includes(text):
    return include_re.findall(text)


def normalize(root, path):
    if posixpath.isabs(path) or posixpath.dirname(path):
        return path

    return posixpath.join(root, path)


def schema_paths(mapper, root_schema):
    unvisited = deque([root_schema])
    visited = []
    while unvisited:
        path = unvisited.popleft()
        visited.append(path)

        text = mapper[path].decode()
        includes = extract_includes(text)

        current_root = posixpath.dirname(path)
        normalized = [normalize(current_root, p) for p in includes]

        unvisited.extend([p for p in normalized if p not in visited])

    return visited


def open_schema(mapper, schema):
    """fsspec-compatible way to open remote schema files

    Parameters
    ----------
    fs : fsspec.filesystem
        pre-instantiated fsspec filesystem instance
    root : str
        URL of the root directory of the schema files
    name : str
        File name of the schema to open.
    glob : str, default: "*.xsd"
        The glob used to find other schema files

    Returns
    -------
    xmlschema.XMLSchema
        The opened schema object
    """
    paths = schema_paths(mapper, schema)
    preprocessed = [remove_includes(mapper[p].decode()) for p in paths]

    return xmlschema.XMLSchema(preprocessed)


def read_xml(mapper, path):
    raw_data = mapper[path]
    tree = etree.fromstring(raw_data)

    namespaces = keymap(lambda x: x if x is not None else "rcm", tree.nsmap)
    schema_location = tree.xpath("./@xsi:schemaLocation", namespaces=namespaces)[0]
    _, schema_path_ = schema_location.split(" ")

    schema_path = posixpath.normpath(
        posixpath.join(posixpath.dirname(path), schema_path_)
    )
    schema = open_schema(mapper, schema_path)

    decoded = schema.decode(tree)

    return decoded



import os
import posixpath
from fnmatch import fnmatchcase

import datatree
import fsspec
import xarray as xr
from fsspec.implementations.dirfs import DirFileSystem
from tlz.dicttoolz import valmap
from tlz.functoolz import compose_left, curry, juxt

from .calibrations import read_noise_levels
from .manifest import read_manifest
from .product.reader import read_product
from .product.transformers import extract_dataset
from .product.utils import starcall
from .xml import read_xml

try:
    ExceptionGroup
except NameError:
    from exceptiongroup import ExceptionGroup


@curry
def execute(tree, f, path):
    node = tree[path]

    return f(node)


def ignored_file(path, ignores):
    ignored = [
        fnmatchcase(path, ignore) or fnmatchcase(posixpath.basename(path), ignore)
        for ignore in ignores
    ]
    return any(ignored)


def open_rcm(
    url,
    *,
    backend_kwargs=None,
    manifest_ignores=[
        "*.pdf",
        "*.html",
        "*.xslt",
        "*.png",
        "*.kml",
        "*.txt",
        "preview/*",
    ],
    **dataset_kwargs,
):
    """read SAFE files of the radarsat constellation mission (RCM)

    Parameters
    ----------
    url : str
    backend_kwargs : mapping
    manifest_ignores : list of str, default: ["*.pdf", "*.html", "*.xslt", "*.png", \
                                              "*.kml", "*.txt", "preview/*"]
        Globs that match files from the manifest that are allowed to be missing.
    **dataset_kwargs
        Keyword arguments forwarded to `xr.open_dataset`, used to open
        the contained data files.
    """
    if not isinstance(url, (str, os.PathLike)):
        raise ValueError(f"cannot deal with object of type {type(url)}: {url}")

    if backend_kwargs is None:
        backend_kwargs = {}

    url = os.fspath(url)

    storage_options = backend_kwargs.get("storage_options", {})
    mapper = fsspec.get_mapper(url, **storage_options)
    relative_fs = DirFileSystem(path=url, fs=mapper.fs)

    try:
        declared_files = read_manifest(mapper, "manifest.safe")
    except (FileNotFoundError, KeyError):
        raise ValueError(
            "cannot find the `manifest.safe` file. Are you sure this is a SAFE dataset?"
        )

    missing_files = [
        path
        for path in declared_files
        if not ignored_file(path, manifest_ignores) and not relative_fs.exists(path)
    ]
    if missing_files:
        raise ExceptionGroup(
            "not all files declared in the manifest are available",
            [ValueError(f"{p} does not exist") for p in missing_files],
        )

    tree = read_product(mapper, "metadata/product.xml")

    calibration_root = "metadata/calibration"
    lookup_table_structure = {
        "/incidenceAngles": {
            "path": "/imageReferenceAttributes",
            "f": compose_left(
                lambda obj: obj.attrs["incidenceAngleFileName"],
                curry(posixpath.join, calibration_root),
                curry(read_xml, mapper),
                curry(extract_dataset, dims="coefficients"),
            ),
        },
        "/lookupTables": {
            "path": "/imageReferenceAttributes/lookupTableFileName",
            "f": compose_left(
                lambda obj: obj.stack(stacked=["sarCalibrationType", "pole"]),
                lambda obj: obj.reset_index("stacked"),
                juxt(
                    compose_left(
                        lambda obj: obj.to_series().to_dict(),
                        curry(valmap, curry(posixpath.join, calibration_root)),
                        curry(valmap, curry(read_xml)(mapper)),
                        curry(valmap, curry(extract_dataset, dims="coefficients")),
                        curry(valmap, lambda ds: ds["gains"].assign_attrs(ds.attrs)),
                        lambda d: xr.concat(list(d.values()), dim="stacked"),
                    ),
                    lambda obj: obj.coords,
                ),
                curry(starcall, lambda arr, coords: arr.assign_coords(coords)),
                lambda arr: arr.set_index({"stacked": ["sarCalibrationType", "pole"]}),
                lambda arr: arr.unstack("stacked"),
                lambda arr: arr.rename("lookup_tables"),
            ),
        },
        "/noiseLevels": {
            "path": "/imageReferenceAttributes/noiseLevelFileName",
            "f": curry(read_noise_levels, mapper, calibration_root),
        },
    }
    calibration = valmap(
        lambda x: execute(**x)(tree),
        lookup_table_structure,
    )

    imagery_paths = tree["/sceneAttributes/ipdf"].to_series().to_dict()
    resolved = valmap(
        compose_left(
            curry(posixpath.join, "metadata"),
            posixpath.normpath,
        ),
        imagery_paths,
    )
    imagery_dss = valmap(
        compose_left(
            curry(relative_fs.open),
            curry(xr.open_dataset, engine="rasterio", **dataset_kwargs),
        ),
        resolved,
    )
    dss = [ds.assign_coords(pole=coord) for coord, ds in imagery_dss.items()]
    imagery = xr.concat(dss, dim="pole")

    return tree.assign(
        {
            "lookupTables": datatree.DataTree.from_dict(calibration),
            "imagery": datatree.DataTree(imagery),
        }
    )



from tlz import filter
from tlz.functoolz import compose_left, curry
from tlz.itertoolz import concat, get

from .product.dicttoolz import query
from .xml import read_xml


def merge_location(loc):
    locator = loc["@locator"]
    href = loc["@href"]

    return f"{locator}/{href}".lstrip("/")


def read_manifest(mapper, path):
    structure = {
        "/dataObjectSection/dataObject": compose_left(
            curry(
                map,
                compose_left(
                    curry(get, "byteStream"),
                    curry(
                        map,
                        compose_left(
                            curry(get, "fileLocation"), curry(map, merge_location)
                        ),
                    ),
                    concat,
                ),
            ),
            concat,
        ),
        "/metadataSection/metadataObject": compose_left(
            curry(
                filter,
                compose_left(curry(get, "@classification"), lambda x: x == "SYNTAX"),
            ),
            curry(map, compose_left(curry(get, "metadataReference"), merge_location)),
        ),
    }

    manifest = read_xml(mapper, path)

    return list(concat(func(query(path, manifest)) for path, func in structure.items()))
